#donotrun
#info DE-RaytracerKn-Pk-1.0.7.frag: Original shader by syntopia. Modifications by knighty + Eiffie + ChrisJRodgers + Patryk Kizny
#info  - Added multiple reflections, multilayer clouds and directional sunlight + point spotlight
#info  - Added color tagging of layers. Include define TAG_LAYERS to enable
#info  - Added depth pass. Include define DEPTH_PASS to enable. Make sure to disable reflections and enable FlatPass.

//  2015/07/22
// Adding background

//  2015/07/15
//  Aded depth pass

// Log Patryk Kizny Modifications
//  2015/07/14
//  Aded color tagging of layers

//  2015/07/12
//  Fixed AO color - now coloredAO color is included for spot and sun

//  2015/06/29
// Specular is now dependent on the intensity of sunlight and spotlight

//  2015/06/20
// - added multilayer clouds
// - added back directional spotlight (sunlight) working together with point spotlight
// - shadows seem to work fine together
// - Todo: shadows can be probably computed faster - did not optimize it

//  2015/06/23
// - added more intuitive camera FOV

//Log:
//mar 7 2015:
// - Now the fog is colored and effects lighting.
// - Colored reflections.
//mar 6 2015:
// - Black dots again. Visible especially when using reflections. Got hard time to fix it. :-/. Any time a transcendental function is involved, one have to take care about infinities, denormals and NaNs.
//mar 5 2015:
// - Implemented stratified sampling for volumetric light and "corrected" use of RNG.
//mar 3 2015:
// -Added Iq's cloud implemented by Eifie. In order to use it #define USE_IQ_CLOUDS before including DE-kn2.frag
// -Added volumetric light. In order to use it #define KN_VOLUMETRIC before including DE-kn2.frag.
//mar 1 2015:
// -Updated settings.
// -Tweaked the AO function to make it "multiscale". Work still in progress: Identify and implement pertinent parameteres. See Baird-Delta preset "AO_Alone"
// -Corrected warnings.
//feb 24 2015:
// -Corrected some bugs. Floating point numbers are tricky.
// -Corrected bug with DoF and Floor.
// -Added point kight distance falloff.
// -Modified height fog formula again.
//feb 23 2015:
// -switched to Eiffie's shoft shadow. Looks much better to me.
// -added altitude fog.
// -ChristRodgers added Point light with glow. Modified a bit.
// -Added bloom effect. Bug corrected.
//Known bugs:
// -Still some black spots on some graphics card. Maybe a driver issue. Don't use unrealistic values for DoF effect plz ;o)
//ToDo:
// -Simplify point light glow computations. Low priority: The light glow is good but suffers from banding artifacts (needs high dither values).
// -Implement user defined matrials. High priority.
// -Reduce banding artifacts ue to tracing accuracy. High priority.
// -SoC? Global illumination? (maybe needs complete refactoring and/or removing InFocusAWidth which may complicate things)

#include "3D-PK-1.0.3.frag"
//#include "3DKn-1.0.1.frag"

#group Raytracer
// Distance to object at which raymarching stops.
uniform float Detail;slider[-7,-3.5,0];
// The step size when sampling AO (set to 0 for old AO)
uniform float DetailAO;slider[-7,-0.5,0];

const float ClarityPower = 1.0;

// Sets focal plane to Target location
uniform bool AutoFocus; checkbox[false]

// Lower this if the system is missing details
uniform float FudgeFactor;slider[0,0.95,1];

float minDist = pow(10.0,Detail);
float aoEps = pow(10.0,DetailAO);
//float MaxDistance = 100.0;

// Maximum number of  raymarching steps.
uniform int MaxRaySteps;  slider[0,256,5000]
uniform float MaxDistance;slider[0,32,1000];
// Use this to boost Ambient Occlusion and Glow
uniform float  MaxRayStepsDiv;  slider[0,1.8,10]

// Can be used to remove banding
uniform float Dither;slider[0,0.5,1];

// Used to prevent normals from being evaluated inside objects.
uniform float NormalBackStep; slider[0,1,10]

#group Depth
uniform float DepthScale; slider[-1.0,0.0,1.0]
uniform float DepthOffset; slider[-5.0,0.0,5.0]
uniform float DepthOffsetPost; slider[-1.0,0.0,1.0]
uniform int DepthScalePower; slider[0,2,16]
uniform bool  MultiChannelDepth; checkbox[false]
uniform bool FlatPass; checkbox[false]

#group Layer Tag
uniform 	vec3 tagBackground; color[0.0,0.0,0.0]
uniform 	vec3 tagLight; color[1.0,0.0,0.0]
uniform 	vec3 tagFloor; color[0.0,1.0,0.0]
uniform 	vec3 tagObject; color[0.0,1.0,1.0]
uniform 	vec3 tagFractal; color[0.0,0.0,1.0]

#group Light

// AO based on the number of raymarching steps
uniform vec4 AO; color[0,0.7,1,0.0,0.0,0.0];
uniform float AoCorrect; slider[0.0,0.0,1.0]
// The specular intensity of the directional light
uniform float Specular; slider[0,0.4,2.0];
// The specular exponent
uniform float SpecularExp; slider[0,16.0,500.0];
// Limits the maximum specular strength to avoid artifacts
uniform float SpecularMax; slider[0,10,100];
//  Color and power of the sunlight
uniform vec4 SunLight; color[0.0,0.4,1.0,1.0,1.0,1.0];
// Direction to the sunlight (spherical coordinates)
uniform vec2 SunDir;  slider[(-1,-1),(0.1,0.1),(1,1)];
// Quality shdows for sunlight only
uniform bool QualityShadows; checkbox[true];

//ToDo - not working with Kn2 mods
uniform bool DebugSun; checkbox[false] Locked

// Light coming from the camera position (diffuse lightning)
uniform vec4 CamLight; color[0,1,2,1.0,1.0,1.0];
// Controls the minimum ambient light, regardless of directionality
uniform float CamLightMin; slider[0.0,0.0,1.0]
// Glow based on distance from fractal
uniform vec4 Glow; color[0,0.0,1,1.0,1.0,1.0];
//uniform vec4 InnerGlow; color[0,0.0,1,1.0,1.0,1.0];
uniform int GlowMax; slider[0,20,1000]

// Adds old-type fog based on distance
uniform float Fog; slider[0,0.0,2]
uniform float FogExp; slider[0,2,10]

uniform vec3 Reflection; color[1,1,1]
uniform int ReflectionsNumber; slider[0,0,5]

vec4 orbitTrap = vec4(10000.0);
float fractionalCount = 0.0;

#group SpotLight
// Color and strength of the spot light
uniform bool SpotGlow; checkbox[true]
uniform vec4 SpotLight; color[0.0,1.,10.0,1.0,1.0,1.0];
uniform vec3 SpotLightPos; slider[(-10,-10,-10),(0,0,0),(10,10,10)]
uniform float LightSize; slider[0.0,0.1,1]
uniform float LightFallOff; slider[0,0,2]
uniform float LightGlowRad; slider[0.0,0.0,5.0]
uniform float LightGlowExp; slider[0.0,1.0,5.0]
// Hard shadows shape is controlled by SpotLightDir + sunlight dir
uniform float HardShadow; slider[0,0,1]
uniform float ShadowSoft; slider[0.0,10.0,20]

// define few vars to trigger shadow computation
int UseSpot = int(SpotLight.w>0);
int UseSun =  int(SunLight.w>0);
bool CastShadows = ((SpotLight.w + SunLight.w) >0);
float sunShadowStrength;

#group Coloring

// This is the pure color of object (in white light)
uniform vec3 BaseColor; color[1.0,1.0,1.0];
// Determines the mix between pure light coloring and pure orbit trap coloring
uniform float OrbitStrength; slider[0,0,1]
// Closest distance to YZ-plane during orbit
uniform vec4 X; color[-1,0.7,1,0.5,0.6,0.6];
// Closest distance to XZ-plane during orbit
uniform vec4 Y; color[-1,0.4,1,1.0,0.6,0.0];
// Closest distance to XY-plane during orbit
uniform vec4 Z; color[-1,0.5,1,0.8,0.78,1.0];
// Closest distance to  origin during orbit
uniform vec4 R; color[-1,0.12,1,0.4,0.7,1.0];
// Background color
uniform vec3 BackgroundColor; color[0.6,0.6,0.45]
// Vignette background
uniform float GradientBackground; slider[0.0,0.3,5.0]

// -----------------------------------------------------------------------------------------------------

float DE(vec3 pos) ; // Must be implemented in other file

uniform bool CycleColors; checkbox[false]
uniform float Cycles; slider[0.1,1.1,32.3]

float DElight(vec3 pos, vec3 lightPos) {
return length(lightPos-pos)-LightSize;
}

float DEL(vec3 pos, inout float lightde, vec3 lightPos) {
lightde=DElight(pos, lightPos)-LightSize;
return min(DE(pos)*FudgeFactor,lightde);
}

float DEL2(vec3 pos, inout float lightde, vec3 lightPos) {
lightde=DElight(pos, lightPos)-LightSize;
return min(DE(pos),lightde);
}


#ifdef providesNormal
vec3 normal(vec3 pos, float normalDistance);

#else
#if 0
vec3 normal(vec3 pos, float normalDistance) {
normalDistance = max(normalDistance*0.5, 1.0e-5);
vec3 e = vec3(0.0,normalDistance,0.0);
vec3 n = vec3(DE(pos+e.yxx)-DE(pos-e.yxx),
	DE(pos+e.xyx)-DE(pos-e.xyx),
	DE(pos+e.xxy)-DE(pos-e.xxy));
n =  normalize(n);
return n;
}
#else
vec3 normal(vec3 pos, float normalDistance) {
normalDistance = max(normalDistance*0.5, 1.0e-5);
vec3 e = vec3(0.0,normalDistance,0.);
vec3 n = vec3(DE(pos+e.yxx)-DE(pos-e.yxx),
	DE(pos+e.xyx)-DE(pos-e.xyx),
	DE(pos+e.xxy)-DE(pos-e.xxy));
n = normalize(n);
return n==n ? n : vec3(0.0);
}
#endif
#endif

#ifdef providesBackground
	vec3  backgroundColor(vec2 dir);
#endif

#group Floor
uniform bool EnableFloor; checkbox[false]
uniform vec3 FloorNormal; slider[(-1,-1,-1),(0,0,1),(1,1,1)]
uniform float FloorHeight; slider[-5,0,5]
uniform vec3 FloorColor; color[1,1,1]

bool floorHit = false;
float floorDist = 0.0;
vec3 floorNormal = normalize(FloorNormal);
float fSteps = 0.0;

#group Height_Fog
uniform float HF_Fallof; slider[0.0005,0.1,5.]
uniform float HF_Const; slider[0.0,0.0,1.0]
uniform float HF_Intensity; slider[0.0,0.0,1.0]
uniform vec3 HF_Dir; slider[(-1,-1,-1),(0,0,1),(1,1,1)]
uniform float HF_Offset; slider[-10.,0.0,10.0]
uniform vec4 HF_Color; color[0.0,1.,3.0,1.0,1.0,1.0]

#ifdef KN_VOLUMETRIC
//to modify volumetric light intensity
uniform float HF_Scatter; slider[0.0,0.,10.0]
//Anisotropy of the fog.
//uniform float HF_Anisotropy; slider[0.0,0.,1.0]
uniform vec3 HF_Anisotropy; color[0,0,0]
//for stratified sampling. Helps to accelerate convergence but slows down the framerate. Better use it for final render. You'll need less subframes.
uniform int HF_FogIter; slider[1,1,16]
//Is shadow cast for every sample. Slows down the rendering.
uniform bool HF_CastShadow; checkbox[false]
#endif

float exp1(float x){ return exp(clamp(x,-80.,80.));}//have to use this to avoid black dots (again)
vec3 exp1(vec3 x){ return exp(clamp(x,vec3(-80.),vec3(80.)));}

float fogAmount(vec3 P0, vec3 P1){//Modified from: http://www.iquilezles.org/www/articles/fog/fog.htm
//ToDo: make dependent on the fog color
vec3 hfdir=normalize(HF_Dir);
float t=length(P1-P0);
float A=HF_Fallof*dot(P1-P0,hfdir);
A=(1.0-exp1(-A))/A;
float amount= (HF_Intensity * exp1(-HF_Fallof*(dot(P0,hfdir)-HF_Offset))* A  + HF_Const) * t;
//amount=clamp(amount,-5.,100.);
return clamp(exp1(-amount),0.,1.);//Return transmission factor
}

vec3 fogAmount3(vec3 P0, vec3 P1){//Modified from: http://www.iquilezles.org/www/articles/fog/fog.htm
//ToDo: make dependent on the fog color
vec3 hfdir=normalize(HF_Dir);
float t=length(P1-P0);
float A=HF_Fallof*dot(P1-P0,hfdir);
A=(1.0-exp1(-A))/A;
vec3 amount= HF_Color.rgb*(HF_Intensity * exp1(-HF_Fallof*(dot(P0,hfdir)-HF_Offset))* A  + HF_Const) * t;
//amount=clamp(amount,-5.,100.);
return clamp(exp1(-amount),vec3(0.),vec3(1.));//Return transmission factor
}

vec3 ptLightGlow(float glow,  vec4 lightColor){
if(SpotGlow){
	float glow1=exp(-pow(LightGlowRad+0.0001,-2.)*pow(glow,LightGlowExp));
	float glow2=exp(-20.*pow(glow,1.));
	return lightColor.xyz*glow2*lightColor.w*1.+ lightColor.xyz*glow1;
}
return vec3(0.);
}

float DEF(vec3 p) {
float d = DE(p) ;
if (EnableFloor) {
	floorDist = abs(dot(floorNormal,p) - FloorHeight);
	if (d<floorDist) {
		fSteps++;
		return d;
	} else return floorDist;
} else {
	fSteps++;
	return d;
}
}

float DEF2(vec3 p) {
if (EnableFloor) {
	floorDist = abs(dot(floorNormal,p)-FloorHeight);
	return min(floorDist, DE(p));
} else {
	return DE(p);
}
}

#define MIN_EPS 2./16777216.

float rand(vec2 co){
// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

#ifdef USE_EIFFIE_SHADOW
// Uses the soft-shadow approach by Eiffie:
float linstep(float a, float b, float t){return clamp((t-a)/(b-a),0.,1.);}
float shadow(vec3 ro, vec3 lightPos, float eps){//float FuzzyShadow(vec3 ro, vec3 rd, float lightDist, float coneGrad, float rCoC){
	//return 1.;
	float rCoC=eps;
	vec3 rd=(lightPos-ro);
	float lightDist=length(rd);
	rd/=lightDist;
	float coneGrad=ShadowSoft;
	if(ShadowSoft==0.) coneGrad=LightSize/lightDist; else coneGrad=1./ShadowSoft;
	float t=DEF2(ro)+rCoC;//avoid self shadowing
	float d=1.0,s=1.0;
	float jitter=Dither*(rand(ro.xy*(float(subframe)+1.0))-0.5);
	for(int i=0;i<MaxRaySteps;i++){
		if(t>lightDist || s<0.001) break;
		float r=rCoC+t*coneGrad;//radius of cone
		#ifndef PERFECT_DE
		//Because most of the time (in particular with fractals) DE is not well defined inside. The soft shadow doesn't "bleed" inside.
		d=DEF2(ro+rd*(t+r*jitter));
		s*=linstep(-r,r,d);//smoothstep(-r,r,d);//
		t+=abs(0.75*d+.25*r);//*mix(1.,0.2*rand(gl_FragCoord.xy*0.02*vec2(i)*t),jitter);
		#else
		//in the case the DE gives correct interior distance this is the "correct" formula giving "correct" shadow.
		d=DEF2(ro+rd*(t+r*jitter))+r;
		s*=linstep(0.,2.*r,1.*d);//smoothstep(-r,r,d);//
		t+=abs(0.75*d+.25*r);
		#endif
	}
	s=max(0.,s-0.001)/0.999;
	return clamp(1.-s,0.0,1.0);
}
#else
// Uses the soft-shadow approach by Quilez:
// http://iquilezles.org/www/articles/rmshadows/rmshadows.htm
float shadow(vec3 ro, vec3 lightPos, float eps) {
	vec3 rd=( lightPos-ro);
	float lightDist=length(rd);
	rd/=lightDist;
	float coneGrad=ShadowSoft;
	if(ShadowSoft==0.) coneGrad=lightDist/(LightSize+0.001); else coneGrad=ShadowSoft;
	float totalDist =2.0*eps;
	float s = 1.0; // where 1.0 means no shadow!
	float jitter=Dither*(rand(ro.xy)-0.5);
	for (int steps=0; steps<MaxRaySteps && totalDist<lightDist; steps++) {
		vec3 p = ro + totalDist*(1.+jitter/coneGrad) * rd;
		float dist = DEF2(p);
		if (dist < eps)  return 1.0;
		s = min(s, coneGrad*(dist/totalDist));
		totalDist += dist;
	}
	return 1.0-s;
}
#endif

// Old type shdows for the sunlight
// Uses the soft-shadow approach by Quilez:
// http://iquilezles.org/www/articles/rmshadows/rmshadows.htm
float sunShadow(vec3 pos, vec3 sdir, float eps) {
	float totalDist =2.0*eps;
	float s = 1.0; // where 1.0 means no shadow!
	for (int steps=0; steps<MaxRaySteps && totalDist<MaxDistance; steps++) {
		vec3 p = pos + totalDist * sdir;
		float dist = DEF2(p);
		if (dist < eps)  return 1.0;
		s = min(s, ShadowSoft*pow((dist/totalDist),1.));
		totalDist += dist;
	}
	return 1.0-s;
}


#if 1
// Ambient occlusion approximation.
// Sample proximity at a few points in the direction of the normal.
// tweaked to make it "multiscale".
float ambientOcclusion(vec3 p, vec3 n) {
	float ao = 0.0;
	float de = DEF(p)/aoEps;//*(-DetailAO);//dividing by aoEps gives control on the overall scale of the AO.
	float wSum = 0.0;
	float w = 1.0;
	float d = 1.0-(Dither*rand(p.xy));
	float D=1.;
	for (float i =1.0; i <15.; i++) {
		// D is the distance estimate difference.
		// If we move 'n' units in the normal direction,
		// we would expect the DE difference to be 'n' larger -
		// unless there is some obstructing geometry in place
		float prevD=D;
		D = (DEF(p+ d*n*de) -0.*de)/(d*de);
		D=min(prevD,D);//if light is obscured at previous level it should be obscured at least at the same amount at this level.
		w *= 1.;//Maybe removed next time
		d*=1.5;//next level  -> x2. maybe becomes a parameter
		ao += w*clamp(1.0-D,0.0,1.0);
		wSum += w;
	}
	return clamp(AO.w*ao/wSum, 0.0, 1.0);
}
#else
// Ambient occlusion approximation.
// Sample proximity at a few points in the direction of the normal.
float ambientOcclusion(vec3 p, vec3 n) {
	float ao = 0.0;
	float de = DEF(p);
	float wSum = 0.0;
	float w = 1.0;
	float d = 1.0-(Dither*rand(p.xy));
	for (float i =1.0; i <10.0; i++) {
		// D is the distance estimate difference.
		// If we move 'n' units in the normal direction,
		// we would expect the DE difference to be 'n' larger -
		// unless there is some obstructing geometry in place
		float D = (DEF(p+ d*n*aoEps) -de)/(d*aoEps);
		w *= 1.;
		d*=2.;
		ao += w*clamp(1.0-D,0.0,1.0);
		wSum += w;
	}
	return clamp(AO.w*ao/wSum, 0.0, 1.0);
}
#endif

vec3 lighting(vec3 n, vec3 color, vec3 pos, vec3 dir, float eps, out float shadowStrength, float ao) {
	shadowStrength = 0.0;
	sunShadowStrength = 0.0;
	
	// Old type of spot light now called sunlight, no falloff
	vec3 sunDir = normalize(vec3(sin(SunDir.x*3.1415)*cos(SunDir.y*3.1415/2.0), sin(SunDir.y*3.1415/2.0)*sin(SunDir.x*3.1415), cos(SunDir.x*3.1415)));
	float nDotSL = max(0.0,dot(n,sunDir));
	vec3 halfVector = normalize(-dir+sunDir);
	float sunDiffuse = nDotSL*SunLight.w;
	
	float hDotN = max(0.,dot(n,halfVector));
	
	// An attempt at Physcical Based Specular Shading:
	// http://renderwonk.com/publications/s2010-shading-course/
	// (Blinn-Phong with Schickl term and physical normalization)
	
	float sunSpecular =((SpecularExp+2.)/8.)*pow(hDotN,SpecularExp)*
	(SpecularExp + (1.-SpecularExp)*pow(1.-hDotN,5.))*
	nDotSL*Specular * SunLight.w;
	
	//  spot light
	float D2L2=dot(SpotLightPos-pos,SpotLightPos-pos);
	float falloff=pow(D2L2,-LightFallOff);
	vec3 spotDir = normalize(SpotLightPos-pos);
	
	float nDotL = max(AoCorrect,dot(n,spotDir));
	halfVector = normalize(-dir+spotDir);
	float diffuse = nDotL;
	
	float ambient = max(CamLightMin,dot(-n, dir));
	hDotN = max(0,dot(n,halfVector));
	
	//  An attempt at Physcical Based Specular Shading:
	//  http://renderwonk.com/publications/s2010-shading-course/
	// (Blinn-Phong with Schickl term and physical normalization)
	// float specular =((SpecularExp+2.)/8.)*pow(hDotN,SpecularExp+0.00001)*(SpecularExp + (1.-SpecularExp)*pow(1.-hDotN,5.))*nDotL*Specular;
	
	float f0=(SpecularExp-1.)/(SpecularExp+1.); f0*=f0;
	float fresnel=f0+(1.-f0)*pow(1.+dot(n,dir),5.);
	float specular =((SpecularExp+2.)/8.)*fresnel*nDotL*pow(hDotN,SpecularExp+0.00001)*Specular * SpotLight.w;
	
	if (HardShadow>0.0 && CastShadows) {
		// check path from pos to spotDir
		shadowStrength = shadow(pos+n*eps, SpotLightPos, eps); // spot
		sunShadowStrength =  sunShadow(pos+n*eps, sunDir, eps); // sun
		
		//shadowStrength = shadow(pos+n*eps, SpotLightPos, eps)*UseSpot + sunShadow(pos+n*eps, sunDir, eps)* UseSun)/(UseSpot + UseSun);
	
	/*float shS= shadow(pos+n*eps, LightPos, 0.1); // attempt for SSS effect. Yes it is possible if the DE is quasi perfect (accurate and defined inside).
	shS=pow(shS,.5);
	shadowStrength = mix(shadowStrength, shS,0.5);*/
	//ambient = mix(ambient,0.0,0.0);
	diffuse = mix(diffuse,0.0,HardShadow*shadowStrength);
	sunDiffuse = mix(sunDiffuse,0.0,HardShadow*sunShadowStrength);
	
	// specular = mix(specular,0.0,HardShadow*f);
	// specular = mix(specular,0.0,HardShadow*shadowStrength); // PK uncomment
	
	specular*=1.-shadowStrength;
	sunSpecular*=1.-sunShadowStrength;
	
	/*
	// Block specular in shadows - turned off because now we have sunlight and spot
	if (shadowStrength>0.0) {specular = 0.0} else {
		specular = mix(specular,0.0,HardShadow*shadowStrength);
		specular*=1.-shadowStrength;
		
	}
	*/
	
};
sunSpecular = min(SpecularMax,sunSpecular);
specular = min(SpecularMax,specular);

// originl Kn2
//vec3 FG=fogAmount3(pos,LightPos);//fog Attenuates light.
//return (FG*SpotLight.xyz*SpotLight.w*falloff*(diffuse+ specular)+CamLight.xyz*CamLight.w*(ambient)*(1.-ao))*color;
// modded for Light in param
//return (FG * lightColor.xyz*lightColor.w*falloff*(diffuse+ specular) + CamLight.xyz*CamLight.w*(ambient)*(1.-ao)*color);

vec3 FG = fogAmount3(pos, SpotLightPos); //fog Attenuates light.

vec3 coloredAO = 1.0 - ao*(1-AO.xyz);
// note: added AO color on spot and sun, originally was only on ambient

//vec3 camLight =  CamLight.xyz*CamLight.w*(ambient)*(1.0 - ao)*color;
vec3 camLight =  CamLight.xyz*CamLight.w*(ambient)*coloredAO*color; // added color to ao
vec3 spotLight = FG*SpotLight.xyz*SpotLight.w*falloff*(diffuse*color*coloredAO + specular);
vec3 sunLight = SunLight.xyz* sunDiffuse*color*coloredAO + SunLight.xyz * sunSpecular;
vec3 totalLight = camLight + spotLight + sunLight;
return totalLight;
}

vec3 colorBase = vec3(0.0,0.0,0.0);

vec3 cycle(vec3 c, float s) {
return vec3(0.5)+0.5*vec3(cos(s*Cycles+c.x),cos(s*Cycles+c.y),cos(s*Cycles+c.z));
}

vec3 getColor() {
orbitTrap.w = sqrt(orbitTrap.w);

vec3 orbitColor;
if (CycleColors) {
	orbitColor = cycle(X.xyz,orbitTrap.x)*X.w*orbitTrap.x +
	cycle(Y.xyz,orbitTrap.y)*Y.w*orbitTrap.y +
	cycle(Z.xyz,orbitTrap.z)*Z.w*orbitTrap.z +
	cycle(R.xyz,orbitTrap.w)*R.w*orbitTrap.w;
} else {
	orbitColor = X.xyz*X.w*orbitTrap.x +
	Y.xyz*Y.w*orbitTrap.y +
	Z.xyz*Z.w*orbitTrap.z +
	R.xyz*R.w*orbitTrap.w;
}

vec3 color = mix(BaseColor, 3.0*orbitColor,  OrbitStrength);
return color;
}

#ifdef  providesColor
vec3 baseColor(vec3 point, vec3 normal);
#endif
float epsModified = 0.0;
bool lighthit = false;

// ------------------------------------------------------ Trace -----------------------------------------

vec3 trace(inout SRay Ray, inout vec3 hitNormal, inout float glow) {
	
	glow=1000.0;
	vec3 hit = SRCurrentPt(Ray); //from+dir*totalDist;
	orbitTrap = vec4(10000.0);
	//vec3 direction = dir;//normalize(dir);
	floorHit = false;
	floorDist = 0.0;
	float dist = 0.0;
	
	int steps;
	colorBase = vec3(0.0,0.0,0.0);
	
	// We will adjust the minimum distance based on the current zoom
	float eps = minDist;
	float lightde=0.;
	
	float epsModified = max(MIN_EPS, Ray.Pos*eps* FudgeFactor);
	orbitTrap = vec4(10000.0);
	vec3 p = SRCurrentPt(Ray);
	float ldist = min(DEF(p) * FudgeFactor, DElight(p,SpotLightPos));
	ldist *= (Dither*rand(Ray.Direction.xy))+(1.0-Dither);
	SRAdvance(Ray, ldist);
	
	for (steps=0; steps<MaxRaySteps; steps++) {
		vec3 p = SRCurrentPt(Ray); //from + totalDist * direction;
		//dist = DEF(p) * FudgeFactor;
		
		dist = lightde = DElight(p, SpotLightPos);//DEL(p, lightde);
		dist = min(dist,DEF(p) * FudgeFactor);
		glow=min(lightde,glow);
		
		SRAdvance(Ray, dist);
		epsModified = max(MIN_EPS, Ray.Pos*eps* FudgeFactor);
		if (dist < epsModified) {
			// move back
			SRAdvance(Ray, dist-2.5*epsModified); //totalDist -= (epsModified-dist);
			//float l = ldist*(epsModifiet-dist)/(ldist-1.*dist); SRAdvance(Ray, -ldist);
			break;
		}
		if (Ray.Pos > MaxDistance) {
			fSteps -= (Ray.Pos-MaxDistance)/dist;
			break;
		}
		//ldist = dist;
	}
	//if(dist==glow) lighthit=true;
	if (EnableFloor && dist ==floorDist*FudgeFactor) floorHit = true;
	vec3 hitColor=vec3(0.);
	float stepFactor = clamp((fSteps)/float(GlowMax),0.0,1.0);
	vec3 backColor = BackgroundColor;
	if (GradientBackground>0.0) {
		float t = length(coord);
		backColor = mix(backColor, vec3(0.0,0.0,0.0), t*GradientBackground);
	}
	
	if (  steps==MaxRaySteps) orbitTrap = vec4(0.0);
	
	float shadowStrength = 0.0;
	
	if ( dist < epsModified) {
		if (dist==lightde) {
			lighthit=true;
			return SpotLight.xyz* SpotLight.w/(LightSize+0.01);
		}
		
		// We hit something, or reached MaxRaySteps
		hit = SRCurrentPt(Ray); //from + totalDist * direction;
		float ao = AO.w*stepFactor ;
		
		if (floorHit) {
			hitNormal = floorNormal;
			if(dot(floorNormal,p)-FloorHeight<0.) //if (dot(hitNormal,Ray.Direction)>0.0)
			hitNormal *=-1.0;
		} else {
			hitNormal= normal(hit-NormalBackStep*epsModified*Ray.Direction, epsModified); // /*normalE*epsModified/eps*/
		}
		
		#ifdef  providesColor
		hitColor = mix(BaseColor,  baseColor(hit,hitNormal),  OrbitStrength);
		#else
		hitColor = getColor();
		#endif
		#ifndef linearGamma
			hitColor = pow(clamp(hitColor,0.0,1.0),vec3(Gamma));
		#endif
		if (DetailAO<0.0) ao = ambientOcclusion(hit, hitNormal);
		if (floorHit) {
			hitColor = FloorColor;
		}
		//hitColor = mix(hitColor, AO.xyz ,ao);
		hitColor= lighting(hitNormal, hitColor,  hit,  Ray.Direction, epsModified, shadowStrength, ao);
		/*
		// OpenGL  GL_EXP2 like fog
		float f = Ray.Pos;//totalDist;
		//hitColor = mix(hitColor, backColor, 1.0-exp(-pow(Fog,4.0)*f*f));
		hitColor = mix(hitColor, backColor, 1.0-exp(-pow(Fog/FogExp,4.0)*pow(f*f,FogExp)));
		*/
		if (floorHit && false) {
			hitColor +=Glow.xyz*stepFactor* Glow.w*(1.0-shadowStrength);
		}
		
	}
	else {
	
/*		vec2 pos = ;
	vec2 pixelsiz=vec2(dFdx(pos.x),dFdy(pos.y));
	vec4 tex = texture2D(frontbuffer, pos);
	vec3 c = tex.xyz/tex.a;
	*/
	#ifdef providesBackground
		vec2 texPos = gl_FragCoord.xy * PixelScale;
		hitColor = backgroundColor(texPos); //dir);
	#else
		hitColor = backColor;
	#endif
	//hitColor +=Glow.xyz*stepFactor* Glow.w*(1.0-shadowStrength);
	
	if (DebugSun) {
		vec3 SunDir =  normalize(vec3(sin(SunDir.x*3.1415)*cos(SunDir.y*3.1415/2.0), sin(SunDir.y*3.1415/2.0)*sin(SunDir.x*3.1415), cos(SunDir.x*3.1415)));
		if (dot(SunDir,normalize(Ray.Direction))>0.9) hitColor= vec3(100.,0.,0.);
	}
	
	hitNormal = vec3(0.0);
	Ray.Pos = MaxDistance;//1000.;//if nothing hit return a big value. This is necesary for the fog.
}
// reflections need to be 0 for this to work
gl_FragDepth = ((1000.0 / (1000.0 - 0.00001)) +
	(1000.0 * 0.00001 / (0.00001 - 1000.0)) /
	clamp(Ray.Pos/length(Ray.Direction), 0.00001, 1000.0));

return hitColor;
}
// ------------------------------------------------- Layer Tagging ----------------------------------

vec3 traceLayer(inout SRay Ray, inout vec3 hitNormal, inout float glow) {
	
	vec3 hit = SRCurrentPt(Ray);
	
	vec3 hitColor = 0.;
	float lightde = 0.;
	float dist = 0.0;
	int steps;
	colorBase = vec3(0.0,0.0,0.0);
	orbitTrap = vec4(10000.0);
	
	lighthit = false;
	floorHit = false;
	floorDist = 0.0;
	
	float eps = minDist;
	float epsModified = max(MIN_EPS, Ray.Pos*eps* FudgeFactor);
	
	vec3 p = SRCurrentPt(Ray);
	float ldist = min(DEF(p) * FudgeFactor, DElight(p,SpotLightPos));
	ldist *= (Dither*rand(Ray.Direction.xy))+(1.0-Dither);
	
	SRAdvance(Ray, ldist);
	
	for (steps=0; steps < MaxRaySteps; steps++) {
		vec3 p = SRCurrentPt(Ray);
		
		//from + totalDist * direction;
		//dist = DEF(p) * FudgeFactor;
		
		dist = lightde = DElight(p, SpotLightPos);//DEL(p, lightde);
		dist = min(dist,DEF(p) * FudgeFactor);
		glow = min(lightde,glow);
		
		SRAdvance(Ray, dist);
		epsModified = max(MIN_EPS, Ray.Pos*eps* FudgeFactor);
		
		if (dist < epsModified) {
			// move back
			SRAdvance(Ray, dist-2.5*epsModified); //totalDist -= (epsModified-dist);
			//float l = ldist*(epsModifiet-dist)/(ldist-1.*dist); SRAdvance(Ray, -ldist);
			break;
		}
		
		if (Ray.Pos > MaxDistance) {
			fSteps -= (Ray.Pos-MaxDistance)/dist;
			break;
		}
		//ldist = dist;
	}
	//if(dist == glow) lighthit = true;
	
	// Floor hit
	//	if (EnableFloor && dist == floorDist * FudgeFactor){	floorHit = true;	}
	if (EnableFloor && dist >= floorDist * FudgeFactor){	floorHit = true;	}
	
	hitColor = 0.;
	float stepFactor = clamp((fSteps)/float(GlowMax),0.0,1.0);
	vec3 backColor = tagBackground;
	
	if (steps == MaxRaySteps) orbitTrap = vec4(0.0);
	float shadowStrength = 0.0;
	
	if ( dist < epsModified) {
		// We hit something, or reached MaxRaySteps
		if (dist == lightde) {
			lighthit = true;
		}
		
		hit = SRCurrentPt(Ray);
		
		if (floorHit){
			hitNormal = floorNormal;
			if(dot(floorNormal,p)-FloorHeight<0.) //if (dot(hitNormal,Ray.Direction)>0.0)
			hitNormal *=-1.0;
			hitColor = tagFloor;
		} else if (lighthit) {
			hitColor = tagLight;
		} else {
			hitColor = tagFractal;
			hitNormal= normal(hit-NormalBackStep*epsModified*Ray.Direction, epsModified); // /*normalE*epsModified/eps*/
		}
		
		#ifndef linearGamma
		//	hitColor = pow(clamp(hitColor,0.0,1.0),vec3(Gamma));
		#endif
	}
	
	else {
		// Background Color
		hitColor = tagBackground;
		
		hitNormal = vec3(0.0);
		Ray.Pos = MaxDistance;//1000.;//if nothing hit return a big value. This is necesary for the fog.
	}
	
	// reflections need to be 0 for this to work
	gl_FragDepth = ((1000.0 / (1000.0 - 0.00001)) +
		(1000.0 * 0.00001 / (0.00001 - 1000.0)) /
		clamp(Ray.Pos/length(Ray.Direction), 0.00001, 1000.0));
	
	return hitColor;
}


// -------------------------------------------------- Depth Pass -----------------------------------------------------------


vec3 splitToChannels(float f){
	vec3 r = 0.0;
	int f24 = int(clamp(f, 0. ,1.) * pow(2,24)); // Cast to 24b size int -> 0 -> 2^24
	//	int f24 = MyTempValue;
	
	r.x = mod(f24, pow(2,24))/pow(2,16); // 24bit comp 0-255 range
	r.y = mod(f24, pow(2,16))/pow(2,8); // 16b comp 0-255 range
	r.z = mod(f24, pow(2,8))/pow(2,0); // 8b comp, 0-255 range
	
	vec3 temp = r.x * pow(2,16) + r.y * pow(2,8) + r.z * pow(2,0);
	//return temp/pow(2,24);
	return r/pow(2,8);
	
}

vec3 traceDepth(inout SRay Ray, inout vec3 hitNormal, inout float glow) {
	bool FlatPass = true;
	vec3 hit = SRCurrentPt(Ray);
	
	vec3 hitColor = 0.;
	float lightde = 0.;
	float dist = 0.0;
	int steps;
	colorBase = vec3(0.0,0.0,0.0);
	
	lighthit = false;
	floorHit = false;
	floorDist = 0.0;
	
	float eps = minDist;
	float epsModified = max(MIN_EPS, Ray.Pos*eps* FudgeFactor);
	
	vec3 p = SRCurrentPt(Ray);
	float ldist = min(DEF(p) * FudgeFactor, DElight(p,SpotLightPos));
	ldist *= (Dither*rand(Ray.Direction.xy))+(1.0-Dither);
	
	SRAdvance(Ray, ldist);
	
	for (steps=0; steps < MaxRaySteps; steps++) {
		vec3 p = SRCurrentPt(Ray);
		
		dist = lightde = DElight(p, SpotLightPos);
		dist = min(dist,DEF(p) * FudgeFactor);
		
		SRAdvance(Ray, dist);
		epsModified = max(MIN_EPS, Ray.Pos*eps* FudgeFactor);
		
		if (dist < epsModified) {
			// move back
			SRAdvance(Ray, dist-2.5*epsModified);
			break;
		}
		
		if (Ray.Pos > MaxDistance) {
			fSteps -= (Ray.Pos-MaxDistance)/dist;
			break;
		}
	}
	
	// Floor hit
	if (EnableFloor && dist == floorDist * FudgeFactor){floorHit = true;}
	
	float stepFactor = clamp((fSteps)/float(GlowMax),0.0,1.0);
	
	vec3 backColor = tagBackground;
	
	if ( dist < epsModified) {
		// We hit something, or reached MaxRaySteps
		if (dist == lightde) {
			lighthit = true;
		}
		
		hit = SRCurrentPt(Ray);
		
		if (floorHit){
			hitNormal = floorNormal;
			if(dot(floorNormal,p)-FloorHeight<0.) hitNormal *=-1.0;
		} else if (lighthit) {
			
		} else {
			
			hitNormal = normal(hit-NormalBackStep*epsModified*Ray.Direction, epsModified);
		}
		
	}
	
	else {
		// Background Color
		hitNormal = vec3(0.0);
		Ray.Pos = MaxDistance;
	}
		
	int dexp = 2;
	float dsmin = 1 / pow(dexp,24);
	float ds = (1-DepthScale) * pow(dexp, DepthScalePower-2);
	
	//float depth = gl_FragDepth = ds - ds/clamp((Ray.Pos + DepthOffset) / length(Ray.Direction), dsmin, MaxDistance) + DepthOffsetPost;
	float depth = gl_FragDepth = clamp(((Ray.Pos + DepthOffset) / length(Ray.Direction))*ds, dsmin, MaxDistance)/MaxDistance + DepthOffsetPost;

	
	vec3 depthColor = 1 - depth;
	if (MultiChannelDepth) depthColor = splitToChannels(1-depth);
	
	return depthColor;
}


// ---------------------------------------------------- Clouds --------------------------------------------

#ifdef USE_IQ_CLOUDS
#group IQ_Clouds
uniform float mtime; slider[-180.00,0,180]

//how large should clouds appear (larger clouds render faster)
uniform float CloudScale; slider[0.1,1.0,10.0]
//how flat are they
uniform float CloudFlatness; slider[0.0,0.0,0.9]
//how high do the clouds go
uniform float CloudTops; slider[-10.0,0.0,10.0]
//and how low
uniform float CloudBase; slider[-10.0,-1.0,10.0]
//how thick
uniform float CloudDensity; slider[0.0,0.0,1.0]
//are they smooth or rough
uniform float CloudRoughness; slider[0.0,1.0,2.0]
//how much light contrast do they create
uniform float CloudContrast; slider[0.0,1.0,10.0]
//what is the base color
uniform vec3 CloudColor; color[0.5,0.5,0.5]
//and the color of light hitting them (posLight doesn't have a color??)
uniform vec3 SunLightColor; color[0.75,0.75,0.75]
//wind direction
uniform vec3 WindDir; slider[(-1.0,-1.0,-1.0),(0.0,1.0,0.0),(1.0,1.0,1.0)]
//wind speed
uniform float WindSpeed; slider[0.0,1.0,2.0]

#ifdef USE_IQ_CLOUDS_MULTILAYER
#group IQ_Clouds High
//how large should clouds appear (larger clouds render faster)
uniform float HCloudScale; slider[0.1,1.0,10.0]
//how flat are they
uniform float HCloudFlatness; slider[0.0,0.5,0.9]
//how high do the clouds go
uniform float HCloudTops; slider[-10.0,2.0,10.0]
//and how low
uniform float HCloudBase; slider[-10.0,1.0,10.0]
//how thick
uniform float HCloudDensity; slider[0.0,0.0,1.0]
//are they smooth or rough
uniform float HCloudRoughness; slider[0.0,1.0,2.0]
//how much light contrast do they create
uniform float HCloudContrast; slider[0.0,1.0,10.0]
//what is the base color
uniform vec3 HCloudColor; color[0.5,0.5,0.5]
//wind direction
uniform vec3 HCloudWindDir; slider[(-1.0,-1.0,-1.0),(0.0,1.0,0.0),(1.0,1.0,1.0)]
//wind speed
uniform float HCloudWindSpeed; slider[0.0,1.0,2.0]
#endif

float rand(vec3 co){
// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
return fract(sin(dot(co*0.123,vec3(12.9898,78.233,112.166))) * 43758.5453);
}

float cnoyz(vec3 co){
vec3 d=smoothstep(0.0,1.0,fract(co));
co=floor(co);
const vec2 v=vec2(1.0,0.0);
return mix(mix(mix(rand(co),rand(co+v.xyy),d.x),
	mix(rand(co+v.yxy),rand(co+v.xxy),d.x),
	d.y),mix(mix(rand(co+v.yyx),rand(co+v.xyx),d.x),
	mix(rand(co+v.yxx),rand(co+v.xxx),d.x),d.y),d.z);
}

float cloudDensity( in vec3 p , in vec3 cloudDir, in float t, float cBase, float cTops, float cDensity, float cScale, float cRoughness, float cFlatness)
{
vec3 q=p/cScale;
q/=vec3(1.0)-cloudDir*cFlatness;
float f=0.0,a=0.5;
for(int i=0;i<5;i++){
f+= a*cnoyz( q );
q = q*2.03;
a = a*0.5;
if(i>6-int(5.0*t/MaxDistance))break;
}
float y=dot(p,cloudDir);
float cDen=1.0 - 2.0*abs(0.5*(cTops+cBase)-y)/(cTops-cBase);
return clamp( cDensity * cDen - cRoughness * f, 0.0, 1.0 );
}

vec4 integrateClouds( in vec4 sum, in float dif, in float den, in float t )
{
// lighting
vec3 lin = CloudColor*1.3 + SunLightColor*dif*CloudContrast; // color from the clouds gui
//vec3 lin = CloudColor*1.3 + SunLight*dif*CloudContrast; // color of the sun

vec4 col = vec4( mix( 1.15*vec3(1.0,0.95,0.8), vec3(0.0), den ), den );
col.xyz *= lin;
col=clamp(col,0.0,1.0);
// col.xyz = mix( col.xyz, bgcol, 1.0-exp(-0.0025*t*t) );
// front to back blending
col.a *= mix(0.9,0.0,clamp(t*t/(MaxDistance*MaxDistance),0.0,1.0));
col.rgb *= col.a;
return sum + col*(1.0-sum.a);
}

vec4 clouds(vec3 p0, vec3 p1,  float cBase, float cTops, float cDensity, float cScale, float cRoughness, float cFlatness, vec3 cWindDir, float cWind){
//vec3 ro=p0 + (WindSpeed*(WindDir*mtime)),rd=normalize(p1-p0),cloudDir=normalize(HF_Dir);
vec3 ro=p0 + (cWind*(cWindDir*mtime)),rd=normalize(p1-p0),cloudDir=normalize(HF_Dir);

vec4 sum = vec4(0.0);
float t=0.1*cScale*rand(vec3(gl_FragCoord.xyy+float(subframe)*30.0)),maxT=length(p1-p0);
bool goingUp=(dot(rd,cloudDir)>0.0);
while(t<maxT) {
vec3  pos = ro + t*rd;
if((!goingUp && dot(pos,cloudDir)<cBase) || (goingUp && dot(pos,cloudDir)>cTops) || sum.a > 0.99) break;
float den = cloudDensity(pos,cloudDir,t, cBase, cTops, cDensity, cScale,  cRoughness, cFlatness);
if( den>0.01 ) {
	vec3 sundir=normalize(SpotLightPos-pos);
	float dif = clamp((den - cloudDensity(pos+0.3*cScale*sundir,cloudDir,t, cBase, cTops, cDensity, cScale,  cRoughness, cFlatness))/0.6, -1.0, 1.0 );
	sum = integrateClouds( sum, dif, den, t );
}
t += 0.1*cScale+0.02*cScale*t;
}
return clamp(sum,0.0,1.0);
}
#endif

#ifdef KN_VOLUMETRIC
float length2(vec3 p){ return dot(p,p);}
//See: https://www.shadertoy.com/view/Xdf3zB
//Good for relatively homogeneous media and for dark ambiance.
//This uses importance sampling wrt distance to light source. Looks good when anisotropy is small but noisy otherwise.
//Should use multipe importance sampling wrt:
// -distance to light (done)
// -Distance: we don't need details far away. (I don't think this is necessary in this particular case)
// -Extinction: when fog density is high enought.
// -Anisotropy: I gess this is as important as distance to light because somehow it have the inverse effect.
//ToDo2: Take fog color into account
//ToDo3: correct the use of pseudo-random number generator. (any help?)
int fogstrata=0;
vec3 ptLightGlow1(vec3 P0, vec3 P1,vec3 lightPos, vec4 lightColor){
//Values used for importance sampling
float A=length2(P0-lightPos);
float B=dot(P1-P0, P0-lightPos);
float C=length2(P1-P0);
float Delta=sqrt(A*C-B*B);
//get sample point using importance sampling.
float x=(rand(viewCoord+vec2(1.6183+float(subframe+fogstrata)))+float(fogstrata))*1./float(HF_FogIter); fogstrata++;
float atanB=atan(B/Delta), atanBC=atan((B+C)/Delta);//B+C=dot((P1-P0)*(P1-LightPos))
float PDF_NF=(atanBC-atanB)/Delta;//Normalization factor due to division by the PDF.
float t=(tan(mix(atanB,atanBC,x))*Delta-B)/C;//when x is close to 0. there are black dots!
t=clamp(t,0.00001,1.); //This is very strange: I have to clamp this way to get rid from black dots. When doing clamp(t,0.,1.) I get much more black dots.
vec3 Pt=P0+t*(P1-P0);//this is the sample point
float extP0t=fogAmount(P0,Pt);//Transmittance between "eye" to sample
float extLt=fogAmount(Pt,lightPos);//Transmittance between sample and light source.
//Compute fog density at sample (redundent with fogAmount(). Should do something about it)
vec3 hfdir=normalize(HF_Dir);
float Ty=HF_Fallof*(dot(Pt,hfdir)-HF_Offset);
float densiTy=HF_Intensity*exp1(-Ty)+HF_Const; //densiTy=clamp(densiTy,0.,100.);
//Shadow
float shadowStrength=1.;
if(HF_CastShadow)
shadowStrength = 1.-shadow(Pt, lightPos, 0.001);
//Anisotopy factor
float cosTheta=dot(normalize(Pt-P0),normalize(lightPos-Pt));//cos(theta): the angle between light direction and view direction at the sample
float denom = 1. + HF_Anisotropy.x*HF_Anisotropy.x - 2.*HF_Anisotropy.x*cosTheta;
float Anie=(1.-HF_Anisotropy.x*HF_Anisotropy.x)/sqrt(denom*denom*denom);//Dropped the 1./(4.*PI)* factor. Well This renderer is not meant to be physically accurate after all. :o)
//Scattering at sample point
float samVal=extP0t*extLt*densiTy*Anie*shadowStrength*sqrt(C);//sqrt(C) is the length of the path.
//Return scattered light
//samVal=max(0.,samVal);
return lightColor.xyz* lightColor.w*PDF_NF*samVal*HF_Scatter;
}

vec3 ptLightGlow3(vec3 P0, vec3 P1){

//Values used for importance sampling
float A=length2(P0-SpotLightPos);
float B=dot(P1-P0, P0-SpotLightPos);
float C=length2(P1-P0);
float Delta=sqrt(A*C-B*B);

//get sample point using importance sampling.
float x=(rand(viewCoord+vec2(1.6183+float(subframe+fogstrata)))+float(fogstrata))*1./float(HF_FogIter); fogstrata++;
float atanB=atan(B/Delta), atanBC=atan((B+C)/Delta);//B+C=dot((P1-P0)*(P1-LightPos))
float PDF_NF=(atanBC-atanB)/Delta;//Normalization factor due to division by the PDF.
float t=(tan(mix(atanB,atanBC,x))*Delta-B)/C;//when x is close to 0. there are black dots!
t=clamp(t,0.00001,1.); //This is very strange: I have to clamp this way to get rid from black dots. When doing clamp(t,0.,1.) I get much more black dots.
vec3 Pt=P0+t*(P1-P0);//this is the sample point
vec3 extP0t=fogAmount3(P0,Pt);//Transmittance between "eye" to sample
vec3 extLt=fogAmount3(Pt,SpotLightPos);//Transmittance between sample and light source.

//Compute fog density at sample (redundent with fogAmount(). Should do something about it)
vec3 hfdir=normalize(HF_Dir);
float Ty=HF_Fallof*(dot(Pt,hfdir)-HF_Offset);
vec3 densiTy=HF_Color.rgb*HF_Intensity*exp1(-Ty)+HF_Const; //densiTy=clamp(densiTy,0.,100.);

//Shadow
float shadowStrength=1.;
if(HF_CastShadow)
shadowStrength = 1.-shadow(Pt, SpotLightPos, 0.001);

//Anisotopy factor
float cosTheta=dot(normalize(Pt-P0),normalize(SpotLightPos-Pt));//cos(theta): the angle between light direction and view direction at the sample
vec3 denom = 1. + HF_Anisotropy*HF_Anisotropy - 2.*HF_Anisotropy*cosTheta;
vec3 Anie=(1.-HF_Anisotropy*HF_Anisotropy)/sqrt(denom*denom*denom);//Dropped the 1./(4.*PI)* factor. Well This renderer is not meant to be physically accurate after all. :o)
//Scattering at sample point
vec3 samVal=extP0t*extLt*densiTy*Anie*shadowStrength*sqrt(C);//sqrt(C) is the length of the path.
//Return scattered light
//samVal=max(0.,samVal);
return SpotLight.xyz*SpotLight.w*PDF_NF*samVal*HF_Scatter;
}
#endif


// ---------------------------------------------------- Color --------------------------------------------


vec3 color(SRay Ray) {

float glow=0.5;
vec3 hitNormal = vec3(0.0);
vec3 col = vec3(0.0);
vec3 col0 = vec3(0.0);
vec3 RWeight=vec3(1);

#ifdef TAG_LAYERS
	return max(vec3(0.),traceLayer(Ray, hitNormal, glow));
#else
#ifdef DEPTH_PASS
return max(vec3(0.),traceDepth(Ray, hitNormal, glow));
#endif

for(int i=0; i<= ReflectionsNumber; i++){
vec3 prevPos = SRCurrentPt(Ray);

col0 = trace(Ray, hitNormal, glow);

vec3 curPos= SRCurrentPt(Ray);
vec3 FG = fogAmount3(prevPos, curPos);//get fog amount
col0=mix(HF_Color.rgb*HF_Color.w,col0,FG);//modify color
col0+=ptLightGlow(max(0.,glow),SpotLight);

#ifdef KN_VOLUMETRIC
if(HF_Scatter*(HF_Intensity+HF_Const)>0.){
vec3 colfogglow=vec3(0.);
for(int j=0;j<HF_FogIter;j++){
colfogglow+=ptLightGlow3(prevPos, curPos);
}
col0+=colfogglow*1./float(HF_FogIter);
//ptLightGlow1(prevPos, curPos);
}
#endif

#ifdef USE_IQ_CLOUDS
//  base layer
vec4 clds = clouds(prevPos, curPos, CloudBase, CloudTops, CloudDensity, CloudScale, CloudRoughness, CloudFlatness, WindDir, WindSpeed);
col0=col0*(1.0-clds.w)+clds.rgb;
#endif

// top layer
#ifdef USE_IQ_CLOUDS_MULTILAYER
clds = clouds(prevPos, curPos, HCloudBase, HCloudTops, HCloudDensity, HCloudScale, HCloudRoughness, HCloudFlatness, HCloudWindDir, HCloudWindSpeed);
col0 = col0*(1.0-clds.w)+clds.rgb;
//ToDo windspeed winddir
#endif

col+=col0*RWeight;
RWeight *= Reflection * FG;//modify current opacity

if (hitNormal == vec3(0.0) || dot(RWeight,RWeight)<0.0001 || lighthit || Ray.Pos>=MaxDistance) {//nothing hit or light hit or reflected light is too small
break;
}

Ray = SRReflect(Ray, hitNormal, 0.*minDist);//reflect the ray
}
return max(vec3(0.),col);//Sometimes col<BigNegativeValue  ->black dots. Why? I don't know :-/. Solved by Eiffie & Syntopia See: http://www.fractalforums.com/fragmentarium/updating-of-de-raytracer/msg81003/#msg81003 . I keep it just in case .
#endif
}
