// Output generated from file: //production/Public/Fractals/Fragmentarium/Frags/Test_Matchmoving.frag
// Created: Cz lip 23 13:27:18 2015
// Based on original classic Quaternion bundled with Fragmentarium
// Modifications added by Patryk Kizny / kizny.com

// 2015/07/12
// - Cleaned up the code and split into frags
// - Added spacetransforms including noise

// 2015/06/15
// - Added Box folding
// - Added Sphere Folding
// - Added Volume clipping
// - Added general fractal repositioning
#info 4D Quaternion Julia DE + Box Folding + Sphere Folding

// Flags
#define providesInit
#define USE_EIFFIE_SHADOW
#define PERFECT_DE


//#define providesBackground
//#define transparentBackground

// Uncomment this to output alpha pass
//#define TAG_LAYERS
// Uncomment this to output depth pass
//#define DEPTH_PASS


//#define USE_IQ_CLOUDS
//#define USE_IQ_CLOUDS_MULTILAYER // Multilayer Clouds by PK
//#define KN_VOLUMETRIC

uniform float time;

// Core
#include "MathUtils.frag"
#include "Quaternion-Maths.frag"
#include "SpaceTransforms.frag"

// Kn1 Renderer
//#include "DE-RaytracerKnPk-1.0.0.frag"
//#include "Presets-Lighting_DE-RaytracerKnPk-1.0.0.frag"
// Kn2 Renderer + Clouds & Stuff
#include "DE-Kn2-PK.1.0.8.frag"


#group 4D Quaternion Julia
// Number of fractal iterations.
uniform int QTIterations;  slider[0,16,100]

// Breakout Distance
uniform float QTThreshold; slider[0,10,100]

// Quaterion Constant C
uniform vec4 QTC; slider[(-1,-1,-1,-1),(0.18,0.88,0.24,0.16),(1,1,1,1)]

// Iterative Rotation
uniform vec3 rotVector; slider[(0,0,0),(0,0,1),(1,1,1)]
uniform float rotAngle; slider[0.00,0,180]

// Iterative Scaling
uniform float Scale;  slider[-3.0,1,3.0]

#group box Folding
// Added box folding by Patryk Kizny
// Sphere folding does not seem to work for this fractal

uniform int foldBoxStartIt; slider[0,0,256]
uniform int foldBoxIts; slider[0,0,16]

uniform vec3 foldBoxLimit; slider[(-5,-5,-5),(0,0,0),(5,5,5)]
uniform vec3 foldBoxValue; slider[(-5,-5,-5),(2,2,2),(5,5,5)]
uniform vec3 foldBoxOffset; slider[(-5,-5,-5),(0.,0.,0.),(5,5,5)]

#group sphere Folding
uniform int foldSphereStartIt; slider[0,0,256]
uniform int foldSphereIts; slider[0,0,16]
uniform float foldSphereMinRad;  slider[0,0.25,2.0]

#group Post
uniform sampler2D Background; file[Ditch-River_2k.hdr]
uniform vec2 bgScale; slider[(0,0),(1,1),(1,1)]
uniform vec2 bgOffset; slider[(0,0),(1,1),(1,1)]

// ---------------------------------------- Init -------------------------------------------

mat3 rot; // QT
void init() {
	rot = rotationMatrix3(normalize(rotVector),rotAngle);
}

// ---------------------------------------- Quaternion Modified -------------------------------------------

vec4 scale = vec4(Scale, Scale, Scale, abs(Scale)) / foldSphereMinRad;

// DE implementation here based on Syntopia + Subblue + Keenan Crane
// http://www.subblue.com/blog/2009/9/20/quaternion_julia
// Added box folding, sphere folding and iterative rotation

float DEQT(vec3 pos) {
	
	vec4 dp = vec4(1.0, 0.0,0.0,0.0);
	vec4 p = vec4(pos, 0.0);

	for (int i = 0; i <QTIterations; i++){
		
		// Quaternion interations
		// Syntopia's compressed version
		//	dp = 2.0* vec4(p.x*dp.x-dot(p.yzw, dp.yzw), p.x*dp.yzw+dp.x*p.yzw+cross(p.yzw, dp.yzw));
		//	p = vec4(p.x*p.x-dot(p.yzw, p.yzw), vec3(2.0*p.x*p.yzw)) + C;
		
		// Experimental, generates artifacts
		//	dp = QTDerCons * qMult(p,dp);
		dp = 2.0 * qMult(p,dp);
		p = qSq(p) + QTC;
		
		// rotation
		p.xyz *= rot;
		
		// scaling, experimental, generates artifacts
		// p *= scale;
		
		// Box Folding (v2) - separate axes + value + offset
		if (foldBoxIts > 0){
			if ( i >= foldBoxStartIt && i < foldBoxStartIt + foldBoxIts){

				p.xyz = clamp(p.xyz, -foldBoxLimit, foldBoxLimit) * foldBoxValue - p.xyz + foldBoxOffset;
			}
		}
		
		// Sphere folding
		if(foldSphereIts > 0){
			if ( i >= foldSphereStartIt && i < foldSphereStartIt + foldSphereIts){
				float r2 = dot(p.xyz, p.xyz);
				p *= clamp(max(foldSphereMinRad/r2, foldSphereMinRad), 0.0, 1.0);
			}
		}
		
		float p2 = dot(p,p);
		orbitTrap = min(orbitTrap, abs(vec4(p.xyz,p2)));
		if (p2 > QTThreshold) break;
	}
	
	// DE
	float r = length(p);
	return  0.5 * r * log(r) / length(dp);
}


// ----------------------------------------Main -------------------------------------------

float DE(vec3 z) {

	// get slice volume
	vec3 z1;
	z1 = opRotate(z, cBoxRotate, length(cBoxRotate));
	z1 = opTranslate(z1, cBoxTranslate);
	float deBox = DEBox(z1, cBoxSize);

	// apply global transformations
	z = opNUScale(z, sScale);
	z = opTranslate(z, sTranslate);
	z = opRotate(z, normalize(sRotate), length(sRotate));
//	z = opTwist(z,sTwist);

	// Stereo projection
	z = stereoMap(z);
	
	// Noise
	z = spaceNoise(z);

	//  Instantiate
	//	z.xz = mod(z.xz,5) -0.0; //+vec2(0.1); //spacingXY
	
	//get fractal
	float deQT = DEQT(z);
 
	// return plain fractal
	float DE = deQT;
	
	// clip to volume
	if (clipToBox){
		DE = max(deQT, deBox);
	}
	// disply cutting volume
	if (showClippingBox) {
		DE = min(deQT, deBox);
	}
	
	return DE;

}

// Returns flat background texture
vec3 backgroundColor(vec2 pos){
	//	sampler2D Background = file("//production/Public/Fractals/Fragmentarium/Camera Test/Frames/Test3/Test_00000.jpg");
		pos = pos * bgScale + bgOffset; 
		return texture2D(Background,pos).xyz;
}



// ---------------------------------------- Presets -------------------------------------------






#preset default
sTranslate = 0,0,0
sRotate = 0,0,0
sScale = 0.7,0.7,0.7
sTwist = 0
sStereo = false
sStereoF = 2
sStereoD1 = 1
sStereoD2 = 2
sStereoOffset = 0,0,0
sStereoMix = 1
clipToBox = true Locked
showClippingBox = false Locked
cBoxSize = 5,5,5
cBoxTranslate = 0,0,5
cBoxRotate = 0,0,0
nTime = 1
nScale = 1,1,1
nAmplitude = 0,0,0
FOV = 1.35
UseTrueFOV = true
TrueFOV = 75
UseFocalLength = false
FocalLength = 14
FilmGate = 36
Eye = 1.22246,-1.65409,0.575885
Ey_ = 0,0,0
Target = 0.606419,-0.0284845,-0.153389
Targe_ = 0,0,0
iCamWeight = 1
iCamScale = 1
iCamPosMod = 1,1,1
iCamDirMod = 1,1,1
iCamPosOffs = 0,0,0
iCamDirOffs = 0,0,0
Frame = 187.975
FocalPlane = 1
Aperture = 0
InFocusAWidth = 0
ApertureNbrSides = 5
ApertureRot = 0
ApStarShaped = false
Gamma = 2.2
ToneMapping = 3
Exposure = 1
Brightness = 1
Contrast = 1
Saturation = 1
GaussianWeight = 1
AntiAliasScale = 2
Bloom = false
BloomIntensity = 0.25
BloomPow = 2
BloomTaps = 4
Detail = -3.5
DetailAO = -0.5
AutoFocus = false
FudgeFactor = 0.95
MaxRaySteps = 768
MaxDistance = 32
MaxRayStepsDiv = 1.8
Dither = 0.5
NormalBackStep = 1
DepthScale = 1
DepthOffset = 0
DepthOffsetPost = 0
DepthScalePower = 0
MultiChannelDepth = false
FlatPass = false
tagBackground = 0,0,0
tagLight = 1,0,0
tagFloor = 0,1,0
tagObject = 0,1,1
tagFractal = 1,1,1
AO = 0,0,0,1
AoCorrect = 0
Specular = 0.02
SpecularExp = 32
SpecularMax = 8
SunLight = 1,1,1,0.3
SunDir = 0.3,0.1
QualityShadows = true
DebugSun = false
CamLight = 1,1,1,0.1
CamLightMin = 0.5
Glow = 1,1,1,0
GlowMax = 256
Fog = 0.35
FogExp = 2
Reflection = 0,0,0
ReflectionsNumber = 0
SpotGlow = false
SpotLight = 1,1,1,0.75
SpotLightPos = 2,0,1
LightSize = 0
LightFallOff = 0
LightGlowRad = 0
LightGlowExp = 1
HardShadow = 1
ShadowSoft = 5
BaseColor = 1,0,0
OrbitStrength = 0.1
X = 0,0.666667,0.498039,0.5
Y = 0,0.666667,1,0.5
Z = 1,0,0.498039,0.5
R = 1,1,1,0.5
BackgroundColor = 0,0,0
GradientBackground = 0.2
CycleColors = false
Cycles = 1
EnableFloor = false
FloorNormal = 0,0,1
FloorHeight = 0
FloorColor = 1,1,1
HF_Fallof = 0.1
HF_Const = 0
HF_Intensity = 0
HF_Dir = 0,0,1
HF_Offset = 0
HF_Color = 1,1,1,1
HF_Scatter = 0
HF_Anisotropy = 0,0,0
HF_FogIter = 1
HF_CastShadow = false
mtime = 0
CloudScale = 1
CloudFlatness = 0
CloudTops = 0
CloudBase = -1
CloudDensity = 0
CloudRoughness = 1
CloudContrast = 1
CloudColor = 0.5,0.5,0.5
SunLightColor = 0.75,0.75,0.75
WindDir = 0,1,0
WindSpeed = 1
HCloudScale = 1
HCloudFlatness = 0.5
HCloudTops = 2
HCloudBase = 1
HCloudDensity = 0
HCloudRoughness = 1
HCloudContrast = 1
HCloudColor = 0.5,0.5,0.5
HCloudWindDir = 0,1,0
HCloudWindSpeed = 1
QTIterations = 16
QTThreshold = 24
QTC = 0.1,0.85,0.1,0
rotVector = 0,1,0
rotAngle = 0
Scale = 1
foldBoxStartIt = 256 Locked
foldBoxIts = 0 Locked
foldBoxLimit = 0,0,0
foldBoxValue = 2,2,2
foldBoxOffset = 0,0,0
foldSphereStartIt = 256 Locked
foldSphereIts = 3 Locked
foldSphereMinRad = 0.25
Background = C:/Users/Paco/Desktop/Fragmentarium 3Dickulus/Examples/Include/Ditch-River_2k.hdr
bgScale = 1,1
bgOffset = 1,1
Up = -0.126412,0.333576,0.850352
#endpreset

